"use strict";
/**
 * Parse files with contents like this:
 *
 *   namespace: component
 *
 *   from .. import
 *     css
 *       dog
 *       cat
 *       bark
 *     id
 *       yes
 *
 *   from /slider import
 *     css
 *       slider
 */
exports.__esModule = true;
var NamespecParser = /** @class */ (function () {
    function NamespecParser(specContents) {
        this.specContents = specContents;
    }
    NamespecParser.prototype.getLine = function () {
        if (this.specContents.length == 0)
            return { eof: true, length: 0 };
        var position = this.specContents.indexOf('\n');
        if (position == -1) {
            return { contents: this.specContents, length: this.specContents.length };
        }
        return { contents: this.specContents.substr(0, position), length: position + 1 };
    };
    NamespecParser.prototype.eof = function () {
        return this.specContents.length == 0;
    };
    NamespecParser.prototype.advance = function (n) {
        this.specContents = this.specContents.substr(n);
    };
    NamespecParser.prototype.consumeWhitespaceWhileExists = function () {
        while (true) {
            var line = this.getLine();
            if (line.contents == null)
                return;
            if (line.contents.match(/^\s*$/)) {
                this.advance(line.length);
            }
            else {
                break;
            }
        }
    };
    NamespecParser.prototype.consumeLine = function (spec, flag) {
        this.consumeWhitespaceWhileExists();
        var line = this.getLine();
        var length = line.length;
        // If at the end of file, return accordingly.
        if (line.eof != null)
            return { eof: true, length: length };
        var contents = line.contents;
        // Try to match the line.
        var match = contents.match(spec);
        // Return early if there is no match.
        if (match == null)
            return { match: false, length: length };
        // Return matching group one otherwise.
        this.advance(length);
        var result = { contents: match[1], allMatches: match, length: length };
        if (flag != null)
            result.flag = flag;
        return result;
    };
    NamespecParser.prototype.consumeNamespace = function () {
        var namespace = this.consumeLine(/^namespace (.+)$/);
        if (namespace.contents == null) {
            throw new Error('Expected to consume namespace');
        }
        return namespace.contents;
    };
    NamespecParser.prototype.consumeImportOrReserve = function () {
        var result = this.consumeLine(/^from ([^ ]+) import$/, 'import');
        if (result.contents == null) {
            result = this.consumeLine(/^(reserve)$/, 'reserve');
        }
        return result;
    };
    NamespecParser.prototype.consumeIndented = function (expectedWhitespace) {
        var type;
        if (expectedWhitespace == null) {
            type = this.consumeLine(/^([ \t]+)([a-zA-Z0-9_-]+)$/);
        }
        else {
            type = this.consumeLine(new RegExp("^(" + expectedWhitespace + ")([a-zA-Z0-9_-]+)$"));
        }
        if (type.allMatches == null)
            return null;
        return {
            value: type.allMatches[2],
            whitespace: type.allMatches[1]
        };
    };
    NamespecParser.prototype.parse = function () {
        var namespace = this.consumeNamespace();
        var expectedTypeWhitespace = null;
        var expectedNameWhitespace = null;
        var imports = new Map();
        var reserves = new Map();
        var result = { namespace: namespace, imports: imports, reserves: reserves };
        while (true) {
            // Iterate through remaining file consuming import statements.
            var entryResult = this.consumeImportOrReserve();
            if (entryResult.contents == null) {
                // If at end-of-file, successful parse!
                this.consumeWhitespaceWhileExists();
                if (this.eof())
                    return result;
                // Throw consume error otherwise.
                throw new Error("Unexpected contents: " + this.specContents.substr(0, 15) + (this.specContents.length > 15 ? '...' : ''));
            }
            var importMap = null;
            var reservedMap = null;
            if (entryResult.flag == 'import') {
                var importNamespace = entryResult.contents;
                if (!result.imports.has(importNamespace)) {
                    // Create the import map.
                    importMap = new Map();
                    result.imports.set(importNamespace, importMap);
                }
                else {
                    importMap = result.imports.get(importNamespace);
                }
            }
            else if (entryResult.flag != 'reserve') {
                throw new Error("Unexpected flag: " + entryResult.flag);
            }
            while (true) {
                // Iterate through import consuming top-level types.
                var type = this.consumeIndented(expectedTypeWhitespace);
                if (type == null)
                    break;
                var typeValue = type.value;
                expectedTypeWhitespace = type.whitespace;
                while (true) {
                    // Iterate through type consuming secondary-level names.
                    var name_1 = void 0;
                    if (expectedNameWhitespace == null) {
                        // If name indent level is unknown, search for more indented than type.
                        name_1 = this.consumeIndented(expectedTypeWhitespace + '[ \t]+');
                    }
                    else {
                        name_1 = this.consumeIndented(expectedNameWhitespace);
                    }
                    if (name_1 == null)
                        break;
                    var nameValue = name_1.value;
                    expectedNameWhitespace = name_1.whitespace;
                    // Add in the type, name pair.
                    if (importMap != null) {
                        // Add import
                        if (importMap.has(typeValue)) {
                            var nameEntries = importMap.get(typeValue);
                            nameEntries.push(nameValue);
                        }
                        else {
                            importMap.set(typeValue, [nameValue]);
                        }
                    }
                    else if (reserves != null) {
                        // Add reserved.
                        if (reserves.has(typeValue)) {
                            var reservedSet = reserves.get(typeValue);
                            reservedSet.add(nameValue);
                        }
                        else {
                            reserves.set(typeValue, new Set([nameValue]));
                        }
                    }
                    else {
                        throw new Error('Expected import or reserved map to be set');
                    }
                }
            }
        }
    };
    return NamespecParser;
}());
exports.NamespecParser = NamespecParser;
